/******************************************************************************
  
  Driver exported entry/exit functions.

  Some of the Linux kernel headers don't like to be compiled with g++ (C++)
  so we need a file that is pure C. This file has two purposes: it implements
  the two Linux module specific entry exit routines, init_module and
  cleanup_module, the file operation callbacks, and also some routines that's
  needed for proc operations.

******************************************************************************/

// This is the only place where this file should be included
// It's unlikely it will be needed anywhere else but if so
// __NO_VERSION__ must be defined before inclusion at the
// other places
#include <linux/module.h>

#include "Lib/OsaIoctl.h"
#include "../OsaRtl.h"

// Headers that cannot be included in C++ code even with macro fixes
// hence we need to provide C functions here for which we give access to the 
// C++ parts using our header (OsaLinuxRtl.h), that is written so that it 
// compiles both in C and C++ (with extern "C" linkage in C++, of course)
#include <linux/proc_fs.h>
#include <asm/uaccess.h>


#include "OsaLinuxEntry.h"

extern int OsaLinuxDriverEntry(void);

static char gs_szOsaProcDirName[] = "osadrivers";
static char gs_szStatusProcFileName[] = "status";

static int          CreateOsaProcDir(void);
static void         RemoveOsaProcDir (void);
static ssize_t      OsaLinuxProcReadCallback(struct file* file, char* buf, size_t count, loff_t *ppos);
static int          OsaLinuxProcPermsCallback (struct inode* inode, int op);
static int          OsaLinuxProcGetInfoCallback(char* buf, char** start, off_t offs, int len, int dummy);
static int          OsaLinuxProcIoControlCallback(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);
static int          OsaLinuxProcOpenCallback(struct inode *inode, struct file *file);
static int          OsaLinuxProcCloseCallback(struct inode *inode, struct file *file);
static unsigned int OsaLinuxPollCallback(struct file *, struct poll_table_struct *);

/******************************************************************************
  Generic /proc/osadrivers/<file> file operations
******************************************************************************/
static struct file_operations gs_sProcFileOperations =
{
  NULL,                           /* lseek   */
  OsaLinuxProcReadCallback,       /* read    */
  NULL,                           /* write   */
  NULL,                           /* readdir */
  OsaLinuxPollCallback,           /* poll    */
  OsaLinuxProcIoControlCallback,  /* ioctl   */
  NULL,                           /* mmap    */
  OsaLinuxProcOpenCallback,       /* open    */
  NULL,                           /* flush   */
  OsaLinuxProcCloseCallback,      /* close   */
  NULL                            /* can't fsync */
};


/******************************************************************************
  Generic /proc/osadrivers/<file> inode operations
******************************************************************************/
static struct inode_operations gs_sProcFileINodeOperations =
{
  &gs_sProcFileOperations,
  NULL,                   /* create */
  NULL,                   /* lookup */
  NULL,                   /* link */
  NULL,                   /* unlink */
  NULL,                   /* symlink */
  NULL,                   /* mkdir */
  NULL,                   /* rmdir */
  NULL,                   /* mknod */
  NULL,                   /* rename */
  NULL,                   /* follow link */
  NULL,                   /* readlink */
  NULL,                   /* readpage */
  NULL,                   /* writepage */
  NULL,                   /* bmap */
  NULL,                   /* truncate */
  OsaLinuxProcPermsCallback
};


/******************************************************************************
  "/proc/osadrivers" node
******************************************************************************/
static struct proc_dir_entry gs_sOsaProcDirNode =
{
  0,                               /* .low_ino */
  sizeof(gs_szOsaProcDirName) - 1, /* .namelen */
  gs_szOsaProcDirName,             /* .name */
  0555 | S_IFDIR,                  /* .mode */
  2,                               /* .nlink */
  0,                               /* .uid */
  0,                               /* .gid */
  0,                               /* .size */
  &proc_dir_inode_operations,      /* .ops */
  NULL,                            /* .get_info */
  NULL,                            /* .fill_inode */
  NULL,                            /* .next */
  NULL,                            /* .parent */
  NULL,                            /* .subdir */
  NULL,                            /* .data */
};

/******************************************************************************
  "/proc/osadrivers/status" node
******************************************************************************/
static struct proc_dir_entry gs_sStatusProcFileNode =
{
  0,                                    /* .low_ino */
  sizeof(gs_szStatusProcFileName) - 1,  /* .namelen */
  gs_szStatusProcFileName,              /* .name */
  0666 | S_IFREG,                       /* .mode */
  1,                                    /* .nlink */
  0,                                    /* .uid */
  0,                                    /* .gid */
  0,                                    /* .size */  // MUST be 0 or we will take this node as an OSA Device
  &gs_sProcFileINodeOperations,         /* .ops */
  &OsaLinuxProcGetInfoCallback,         /* .get_info */
  NULL,                                 /* .fill_inode */
  NULL,                                 /* .next */
  NULL,                                 /* .parent */
  NULL,                                 /* .subdir */
  NULL,                                 /* .data */
};



/******************************************************************************
  Driver main exported entry function.
  Called when driver loads
******************************************************************************/
int init_module(void)         
{
  int rc;
  OSA_DPRINTF(("OsaLinux: Loading Osa based driver\n"));
                              
  // Register the driver as a "Char Device" driver, use dynamic major number assignment
  CreateOsaProcDir();

  OSA_DPRINTF(("OsaLinux: Created proc entries\n"));
  rc = OsaLinuxDriverEntry();
  return rc;   // Something wrong
}


/******************************************************************************
  Driver main exported exit function
  Called when driver unloads
******************************************************************************/
void cleanup_module(void)         
{
  OSA_DPRINTF(("OsaLinux: Unloading Otx based driver\n"));
  RemoveOsaProcDir();
  OsaLinuxUnloadEntry();
  return;
}


/******************************************************************************
  Create a new device node entry (file) in dir "/proc/osadrivers"  
******************************************************************************/
struct proc_dir_entry *OsaLinuxCreateProcDevice(const wchar_t *pwcsDeviceName, void *pContext)
{
  struct proc_dir_entry *de;

  int cbSize = sizeof(char) * (wcslen(pwcsDeviceName) + 1);
  char *pszName = kmalloc(cbSize, GFP_KERNEL);

  de = kmalloc(sizeof(struct proc_dir_entry), GFP_KERNEL);

  memset(de, 0, sizeof (struct proc_dir_entry));

  wcsztosz(pszName, pwcsDeviceName);

  OSA_DPRINTF(("OsaLinux: Name of proc entry = %s\n", pszName));

  OSA_DPRINTF(("OsaLinux: Proc registering %s, pExternalDeviceP=0x%x\n", pszName, (int) pContext));

  de->low_ino = 0;
  de->namelen = strlen(pszName);
  de->name = pszName;
  de->mode = 0666 | S_IFREG;
  de->nlink = 1;
  de->uid = 0;
  de->gid = 0;
  de->size = (int) pContext;
  de->ops = &gs_sProcFileINodeOperations;
  de->get_info = &OsaLinuxProcGetInfoCallback;
  de->fill_inode = NULL;
  de->next = NULL;
  de->parent = NULL;
  de->subdir = NULL;

  proc_register(&gs_sOsaProcDirNode, de);

  return de;
} 


/******************************************************************************
  Remove the device node entry (file) in dir "/proc/osadrivers"
******************************************************************************/
void OsaLinuxRemoveProcDevice(struct proc_dir_entry *pProcEntry)
{
  OSA_ASSERT(pProcEntry);
  if (pProcEntry) {
    proc_unregister(&gs_sOsaProcDirNode, pProcEntry->low_ino);
    OSA_DPRINTF(("OsaLinux: Unregistering proc entry %s \n", pProcEntry->name));
    kfree(pProcEntry->name);
    kfree(pProcEntry);
  }
}


/******************************************************************************
  Create the /proc/osadriver directory
******************************************************************************/
static int CreateOsaProcDir(void)
{
  int err;
  OSA_DPRINTF(("OsaLinux: Registering ProcDirNode\n"));
  err = proc_register(&proc_root, &gs_sOsaProcDirNode);
  
  if (!err) {
    OSA_DPRINTF(("OsaLinux: Registering StatusProcFileNode\n"));
    proc_register(&gs_sOsaProcDirNode, &gs_sStatusProcFileNode);
  }
  return err;
}


/******************************************************************************
  Remove the /proc/osadriver directory and all devices (files) in it
******************************************************************************/
static void RemoveOsaProcDir(void)
{
  struct proc_dir_entry *pOld = NULL;

  proc_unregister(&gs_sOsaProcDirNode, gs_sStatusProcFileNode.low_ino);
  
  // Remove all devices in the directory
  while (gs_sOsaProcDirNode.subdir && (gs_sOsaProcDirNode.subdir != pOld)) {
    pOld = gs_sOsaProcDirNode.subdir;
    if (gs_sOsaProcDirNode.subdir->size) {
      // A proc entry which has a corresponding OSA device (ExternalDevice)
      // The procentry will be removed when the ExternalDevice is removed
      OsaLinuxRemoveDeviceEntry((void *)gs_sOsaProcDirNode.subdir->size);
    } else {
      // Not at proc entry corresponding to an OSA device
      // Since we unregister the status node first we should never get here
      OsaLinuxRemoveProcDevice(gs_sOsaProcDirNode.subdir);
    }
  }
  
  proc_unregister(&proc_root, gs_sOsaProcDirNode.low_ino);
}


/******************************************************************************
  Callback, when reading a proc file node (device), for example with the 'cat'
  command
******************************************************************************/
static ssize_t OsaLinuxProcReadCallback(struct file* file, char* buf, size_t count,
				  loff_t *ppos)
{
  OSA_DPRINTF(("OsaLinux: OsaLinuxProcReadCallback\n"));
  return 0;
}


/******************************************************************************
  Callback
******************************************************************************/
static int OsaLinuxProcPermsCallback (struct inode* inode, int op)
{
  OSA_DPRINTF(("OsaLinux: OsaLinuxProcPermsCallback\n"));
  return 0;
}


/******************************************************************************
  Callback
******************************************************************************/
static int OsaLinuxProcGetInfoCallback(char* buf, char** start, off_t offs, int len,
        int dummy)
{
  int cnt;// = sizeof(conf_hdr) - 1;
  OSA_DPRINTF(("OsaLinux: OsaLinuxProcGetInfoCallback\n"));
  strcpy (buf, "Test");
  cnt = strlen (buf);
  return cnt;
}


/******************************************************************************
******************************************************************************/
int OsaLinuxProcOpenCallback(struct inode *inode, struct file *pFile)
{
  struct proc_dir_entry *pde = pFile->f_dentry->d_inode->u.generic_ip;
  OSA_ASSERT(pde);
  if (pde) {
    return OsaLinuxOpenDeviceEntry(pFile, (void *)pde->size); 
  }
  return -1;
}


/******************************************************************************
******************************************************************************/
int OsaLinuxProcCloseCallback(struct inode *inode, struct file *pFile)
{
  struct proc_dir_entry *pde = pFile->f_dentry->d_inode->u.generic_ip;
  OSA_ASSERT(pde);
  if (pde) {
    return OsaLinuxCloseDeviceEntry(pFile, (void *)pde->size); 
  }
  return -1;
}


/******************************************************************************
  Callback, when user ioctl is called for a device (file) in 
  "/proc/osadrivers/"
******************************************************************************/
int OsaLinuxProcIoControlCallback(struct inode *inode, struct file *pFile,
                unsigned int cmd, unsigned long arg)
{
  struct proc_dir_entry *pde = inode->u.generic_ip;
  OsaLinuxIoCtlHdr sHdr;
  void *pSystemBuffer = NULL;
  int cbSystemBufferSize;

// Note: VERIFY_WRITE is a superset of VERIFY_READ
//  if (access_ok(VERIFY_WRITE, (void *) arg, sizeof(OsaLinuxIoCtlHdr)) {

  //OSA_DPRINTF(("OsaLinux: In OsaLinuxProcIoControlCallback, arg = %lx\n", arg));

  // First get the Linux specific header so we can get to the generic parameters
  if (arg) {
    if (copy_from_user(&sHdr, (void *) arg, sizeof(OsaLinuxIoCtlHdr))) { 
      OSA_EPRINTF(("OsaLinux: Unexpected copy_from_user() error (sHdr)\n"));
      return -EFAULT;
    }
  } else {
    // Lets make it possible for us to handle NULL arguments, although NULL arg this is abnormal
    memset(&sHdr, 0, sizeof(OsaLinuxIoCtlHdr));
  }

  //OSA_DPRINTF(("OsaLinux: sHdr.pOutBuf = %p\n", sHdr.pOutBuf));

  // Allocate space for the input and output buffers
  cbSystemBufferSize = OSA_MAX(sHdr.cbInBufSize, sHdr.cbOutBufSize); 
  if (cbSystemBufferSize) {
    pSystemBuffer = kmalloc(cbSystemBufferSize, GFP_KERNEL);
    if (!pSystemBuffer) {
      OSA_EPRINTF(("OsaLinux: Unexpected kmalloc() error\n"));
      return -EFAULT;
    }
  } else {
    pSystemBuffer = NULL;
  }

  if (sHdr.cbInBufSize) {
    if (copy_from_user(pSystemBuffer, sHdr.pInBuf, sHdr.cbInBufSize)) { 
      OSA_EPRINTF(("OsaLinux: Unexpected copy_from_user() error (sHdr.pInBuf)\n"));
      kfree(pSystemBuffer);
      return -EFAULT;
    }
#if 0
/*
    OSA_DPRINTF(("received system buffer %d bytes in %d out\n", sHdr.cbInBufSize, sHdr.cbOutBufSize));
    {
      int i;
      for (i = 0; i < sHdr.cbInBufSize; i++) {
        OSA_DPRINTF(("%.2x ", ((OsaUint8 *)pSystemBuffer)[i]));
      }
      OSA_DPRINTF(("\n"));
    }
*/
#endif
  }
    
  if (OsaLinuxIoDeviceControlEntry(pFile, (void *)pde->size, cmd, &sHdr, pSystemBuffer)) {
    OSA_EPRINTF(("OsaLinux: OsaLinuxIoDeviceControlEntry failed\n"));
    if (copy_to_user((void *) arg, &sHdr, sizeof(OsaLinuxIoCtlHdr))) { 
      OSA_EPRINTF(("OsaLinux: Unexpected copy_to_user() error (sHdr)\n"));
      return -EFAULT;
    }
    kfree(pSystemBuffer);
    return -EFAULT;
  }

  if (sHdr.cbReturned) {
    OSA_ASSERT(sHdr.cbReturned <= cbSystemBufferSize);
    //OSA_DPRINTF(("OsaLinux: Returning %d bytes in sHdr.pOutBuf\n", sHdr.cbReturned));
    if (sHdr.cbReturned <= cbSystemBufferSize) {
      if (copy_to_user(sHdr.pOutBuf, pSystemBuffer, sHdr.cbReturned) != 0) {
        OSA_EPRINTF(("OsaLinux: Unexpected copy_to_user() error (pSystemBuffer)\n"));
        kfree(pSystemBuffer);
        return -EFAULT;
      } else {
        //OSA_DPRINTF(("Copied %d bytes to user mode from pSystemBuffer\n", sHdr.cbReturned)); 
        if (copy_to_user((void *) arg, &sHdr, sizeof(OsaLinuxIoCtlHdr))) { 
          OSA_EPRINTF(("OsaLinux: Unexpected copy_to_user() error (sHdr)\n"));
          return -EFAULT;
        }
        //OSA_DPRINTF(("Copied header, %d bytes to user mode\n", sizeof(OsaLinuxIoCtlHdr))); 
        //OSA_DPRINTF(("OsaLinux: <1>Copied Hello to user mode\n"));
      }
    } else {
      return -EFAULT;
    }
  }

  if (pSystemBuffer) {
    kfree(pSystemBuffer);
  }
  return 0;
}


/******************************************************************************
  Callback, when user calls poll to ask (and possibly wait) if the device
  is ready for action. i.e. if the device will return or accept data, and do
  so without blocking, if read or written to.
  "/proc/osadrivers/"
******************************************************************************/
static unsigned int OsaLinuxPollCallback(struct file *pFile, struct poll_table_struct *pPollTable)
{
  struct proc_dir_entry *pde = pFile->f_dentry->d_inode->u.generic_ip;
  OSA_ASSERT(pde);
  if (pde) {
    return OsaLinuxPollDeviceEntry(pFile, (void *)pde->size, pPollTable);
  }
  return -1;
}






